#include <cstring>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <string>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <iconv.h>
#include <wchar.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include "../lh_w/common.h"


const int SELECTION_INTERVAL=200; 		    // check selection interval.
int client_sock = -1;
int last_msg = OP_CLOSEWINDOW;
class Selection;
Selection *g_sel = NULL;
GdkModifierType key_mask = GdkModifierType(0);

int run_client();

static bool bIsPureEnglish(const char *str) 
{ 
	for (int i=0; str[i]!=0; i++) 
		if (!isascii(str[i]))
			return false;
	return true;	
}


class Selection {
private:
    gint IsBusy;
	gint timeout;

	void create_selection_widget();

	static gint TimeOutCallback(gpointer data);
	static void SelectionReceivedCallback(GtkWidget* widget,GtkSelectionData *selection_data, guint time, Selection *oSelection);
	void SelectionReceived(gchar* sValue);
	gboolean Enabled();
public:
	GdkAtom UTF8_STRING_Atom, COMPOUND_TEXT_Atom; //need to set it to static? this make oTextWin can't use it conveniently.
	std::string LastClipWord;
	GtkWidget* selection_widget;

	Selection();
	void Init();
	void End();
	void start();
	void stop();
};


Selection::Selection()
{
	UTF8_STRING_Atom = COMPOUND_TEXT_Atom = GDK_NONE;
	IsBusy = 0;
	selection_widget = NULL;
	timeout = 0;
}

/********************************************************************/
gboolean Selection::Enabled()
{
  GdkScreen *screen = gtk_window_get_screen(GTK_WINDOW(selection_widget));
  GdkDisplay *display = gdk_screen_get_display(screen);
  GdkModifierType mask;
  gdk_display_get_pointer(display, NULL, NULL, NULL, &mask);
  bool pressed=false;
  /*
    typedef enum
    {
    GDK_SHIFT_MASK    = 1 << 0,
    GDK_LOCK_MASK     = 1 << 1,
    GDK_CONTROL_MASK  = 1 << 2,
    GDK_MOD1_MASK     = 1 << 3,
    GDK_MOD2_MASK     = 1 << 4,
    GDK_MOD3_MASK     = 1 << 5,
    GDK_MOD4_MASK     = 1 << 6,
    GDK_MOD5_MASK     = 1 << 7,
    GDK_BUTTON1_MASK  = 1 << 8,
    GDK_BUTTON2_MASK  = 1 << 9,
    GDK_BUTTON3_MASK  = 1 << 10,
    GDK_BUTTON4_MASK  = 1 << 11,
    GDK_BUTTON5_MASK  = 1 << 12,
  */

  mask = GdkModifierType(guint(mask) & (1<<13)-1);

  mask = 
    GdkModifierType(guint(mask) & ~(GDK_LOCK_MASK | GDK_BUTTON1_MASK | GDK_BUTTON2_MASK |
				    GDK_BUTTON3_MASK | GDK_BUTTON4_MASK | GDK_BUTTON5_MASK |
				    GDK_MOD2_MASK | GDK_MOD3_MASK | GDK_MOD5_MASK));

  if (mask == key_mask)//comb_mask)
    pressed=true;

	//printf("screen=%x, display=%x, mask=%x, pressed = %d\n", (unsigned int)screen, (unsigned int)display, mask, pressed);
	//printf("mask=%x, pressed = %d\n", mask,   pressed);
	return pressed;
}

void Selection::create_selection_widget()
{
	if (selection_widget)
		gtk_widget_destroy(selection_widget);
	selection_widget = gtk_window_new(GTK_WINDOW_TOPLEVEL);//gtk_invisible_new ();
	gtk_widget_hide(selection_widget);
	//printf("selection_widget=%x\n", selection_widget);
	g_signal_connect (G_OBJECT (selection_widget), "selection_received", G_CALLBACK (SelectionReceivedCallback), this);
}


void Selection::Init()
{
	UTF8_STRING_Atom = gdk_atom_intern("UTF8_STRING",false);
	COMPOUND_TEXT_Atom = gdk_atom_intern("COMPOUND_TEXT",false);

	create_selection_widget();
}

void Selection::End()
{
	stop();
	if (selection_widget)
		gtk_widget_destroy(selection_widget);
}

void Selection::start()
{
	if (!timeout)
		timeout = g_timeout_add(SELECTION_INTERVAL,TimeOutCallback,this);
}

void Selection::stop()
{
	if (timeout) {
		g_source_remove(timeout);
		timeout = 0;
	}
	LastClipWord.clear();
}

gint Selection::TimeOutCallback(gpointer data)
{
    Selection *oSelection = (Selection *)data;

	//printf("TimeOutCallback IsBusy=%d\n", oSelection->IsBusy);
    if (!oSelection->Enabled()) {

		//if (conf->get_bool_at("dictionary/only_scan_while_modifier_key")) {
    	//bool do_scan = gpAppFrame->unlock_keys->is_pressed();
    	//if (!do_scan)
		return true;
    }

    if (oSelection->IsBusy) {
	    oSelection->IsBusy++;
	    if (oSelection->IsBusy*SELECTION_INTERVAL > 8000 ) {
				//"selection_received" signal is not received for 8 seconds.
				//there encounter some error,and i find that this often take a long time (serveral minutes) to get the signal at last,
				//during this perioed,if you call gtk_selection_convert(),the "selection_received" signal will not be received also,
				//and at last these signals are received at almost the same time...BAD.
				
				//so here create a new selection_widget, then call gtk_selection_convert()...this should can throw that error selection.
				//!!!:
				//But this seems(i am not sure) will make the widgets in StarDict become unselectable! see BUGS.
				g_warning("Error, selection data didn't received, retring!\n");
				oSelection->create_selection_widget();
				oSelection->IsBusy = 0;
				if(last_msg != OP_CLOSEWINDOW && client_sock != -1)
				{
					//printf("sending close window message\n");
					short msg = htons(OP_CLOSEWINDOW);
					if(send(client_sock, &msg, 2, 0) <= 0)
					{
						printf("send failed\n");					
						if(!run_client())
						{
							printf("can not launch lh_w.exe.\n");
							if(g_sel)
							{
								delete g_sel;
								g_sel = NULL;
							}
							exit(1);
						}
					}
					last_msg = OP_CLOSEWINDOW;
				}
	    }
	}
	else {
    	oSelection->IsBusy = 1;
    	gtk_selection_convert (oSelection->selection_widget, GDK_SELECTION_PRIMARY, oSelection->UTF8_STRING_Atom, GDK_CURRENT_TIME);
    }

	return true;
}

void Selection::SelectionReceivedCallback(GtkWidget* widget,GtkSelectionData *selection_data, guint time, Selection *oSelection)
{	
	gchar *result;
	result = (gchar *)gtk_selection_data_get_text (selection_data);
	//printf("SelectionReceivedCallback result=%s\n", result);
	if (!result) {
		/* If we asked for UTF8 and didn't get it, try compound_text;
		* if we asked for compound_text and didn't get it, try string;
		* If we asked for anything else and didn't get it, give up.
		*/
		if (selection_data->target == oSelection->UTF8_STRING_Atom) {
			gtk_selection_convert (widget, GDK_SELECTION_PRIMARY, oSelection->COMPOUND_TEXT_Atom, GDK_CURRENT_TIME);
		}
		else if (selection_data->target == oSelection->COMPOUND_TEXT_Atom)
		{
			gtk_selection_convert (widget, GDK_SELECTION_PRIMARY, GDK_TARGET_STRING, GDK_CURRENT_TIME);
		}
		else {
			oSelection->IsBusy = 0;
			oSelection->LastClipWord.clear();
			if(last_msg != OP_CLOSEWINDOW && client_sock != -1)
			{
				//printf("sending close window message\n");
				short msg = htons(OP_CLOSEWINDOW);
				if(send(client_sock, &msg, 2, 0) <= 0)
				{
					printf("send failed\n");					
					if(!run_client())
					{
						printf("can not launch lh_w.exe.\n");
						if(g_sel)
						{
							delete g_sel;
							g_sel = NULL;
						}
						exit(1);
					}
				}
				last_msg = OP_CLOSEWINDOW;
			}
		}
		return;
    }
	oSelection->IsBusy = 0;

	oSelection->SelectionReceived(result);	
	g_free (result);
}

/********************************************************************/
void Selection::SelectionReceived(gchar* sToken)
{
	static int time = 0;
	static int sent = 0;
	
	//static int index = 0;
	//index++;
	//printf("SelectionReceived, index=%d, sToken[0]=%c\n", index, sToken[0]);
    if(sToken[0] == '\0') {
        LastClipWord.clear();

		if(last_msg != OP_CLOSEWINDOW && client_sock != -1)
		{
			//printf("sending close window message\n");
			short msg = htons(OP_CLOSEWINDOW);
			if(send(client_sock, &msg, 2, 0) <= 0)
			{
				printf("send failed\n");					
				if(!run_client())
				{
					printf("can not launch lh_w.exe.\n");
					if(g_sel)
					{
						delete g_sel;
						g_sel = NULL;
					}
					exit(1);
				}
			}
			last_msg = OP_CLOSEWINDOW;
		}
        return;
    }

	gint len = 0;
	gchar *a = sToken;
	while ((*a) && len < 256) {
		a = g_utf8_next_char (a);
		++len;
	}
	*a = '\0';
	
    while (g_ascii_isspace(*sToken))
		sToken++;
	if(sToken[0] == '\0') {
		LastClipWord.clear();

		if(last_msg != OP_CLOSEWINDOW && client_sock != -1)
		{
			//printf("sending close window message\n");
			short msg = htons(OP_CLOSEWINDOW);
			if(send(client_sock, &msg, 2, 0) <= 0)
			{
				printf("send failed\n");					
				if(!run_client())
				{
					printf("can not launch lh_w.exe.\n");
					if(g_sel)
					{
						delete g_sel;
						g_sel = NULL;
					}
					exit(1);
				}
			}
			last_msg = OP_CLOSEWINDOW;
		}
        return;
    }

	a=strchr(sToken,'\n');
	if (a)
		*a='\0';
	

	
    if (g_str_has_prefix(sToken,"http://") || g_str_has_prefix(sToken,"ftp://")) {
        LastClipWord.clear();
        return;
    }
    
	if (LastClipWord != sToken) {	// not equal
		LastClipWord = sToken;
		time = 0;
		sent = 0;
	}
	else if(!sent)
	{
		time += 200;
		if(time > 500)
		{
			sent = 1;
			//printf("%s\n", sToken);
			if(client_sock != -1)
			{
				char buf[2048];
				int count;

				*((short*)buf) = htons(OP_TOKEN);

				size_t in = strlen(sToken);
				size_t out = 2046;
				iconv_t cv = iconv_open("UCS-2", "UTF-8");
				char *d = &buf[2];
				iconv(cv, &sToken, &in, &d, &out);
				iconv_close(cv);
				//printf("out=%d\n", (int)out);
				if(out == 0)
					count = 2048;
				else
					count = 2+2046-out+2;		// op + token + null
				buf[count-1] = buf[count-2] = 0; 	// set the last character to '\0'
				//for(int i = 0; i < count; i++)
				//	printf("0x%02x ", buf[i]);
				//printf("\n");
		
				if(send(client_sock, buf, count, 0) <= 0)
				{
					printf("send failed\n");					
					if(!run_client())
					{
						printf("can not launch lh_w.exe.\n");
						if(g_sel)
						{
							delete g_sel;
							g_sel = NULL;
						}
						exit(1);
					}
				}
				last_msg = OP_TOKEN;

				//FILE *f = fopen("test.txt", "w");
				//fwrite(buf, 1, wcslen(buf)*2+4, f);
				//fclose(f);
				//for(int i=0; i++; buf[i]!=0)
				//	printf("%02x", buf[i]);
			}
		}
	}
}

int run_client()
{
	char buf[256];
	strcpy(buf, getenv("PWD"));
	strcat(buf, "/lh_w.exe");
	//printf("lh_w: %s\n", buf);
	struct stat st;
	if(stat(buf, &st) < 0)
	{
		printf("can not find lh_w.exe!");
		return 0;
	}

	pid_t pid;
	pid = fork();
	if (pid < 0)
	{
		printf("fork error!\n");
		return 0;
	}
	else if (pid == 0) // child
	{
		if(client_sock > 0)
			close(client_sock);
		delete g_sel;
		//memmove(buf+5, buf, strlen(buf)+1);
		//memcpy(buf, "wine ", 5);
		//usleep(1000000);
		if(execlp("wine", "wine", buf, NULL) < 0)
			printf("run lh_w.exe error!\n");
		exit(1);
	}
	else // parent
	{
		socklen_t size;
		struct sockaddr_in service, clientaddr;
	
		int server_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
		if(server_sock < 0)
		{
			printf("can not create socket\n");
			return 0;
		}
		int on = 1;
      if (setsockopt(server_sock, SOL_SOCKET, SO_REUSEADDR, (char*)&on, sizeof(on)) < 0)
        {
			printf("can not set socket option\n");
			return 0;
        }
		service.sin_family = AF_INET;
		service.sin_addr.s_addr = inet_addr("127.0.0.1");
		service.sin_port = htons(SERVER_PORT);
		if (bind(server_sock, (struct sockaddr *) &service, sizeof(service)) < 0) 
		{
			printf("bind failed.\n");
			close(server_sock);
			server_sock = -1;
			return 0;
		}

		if (listen(server_sock, 1 ) < 0)
		{
			printf("listen failed.\n");
			close(server_sock);
			server_sock = -1;
			return 0;
		}

		if(client_sock > 0)
			close(client_sock);
		client_sock = -1;
		size = sizeof(struct sockaddr_in);
		client_sock = accept(server_sock,  (struct sockaddr *) &clientaddr, &size);
		if(client_sock < 0)
		{
			printf("accept failed.\n");
			close(server_sock);
			server_sock = -1;
			return 0;
		}
		shutdown(server_sock, SHUT_RDWR);
		close(server_sock);
		server_sock = -1;
		
		return 1;
	}
}

void sig(int s)
{
	//printf("sig %d\n", s);
	if(s == SIGINT || s == SIGTERM)
	{
		if(g_sel)
		{
			delete g_sel;
			g_sel = NULL;
		}
		//printf("sig %d, clientsock %d\n", s, client_sock);
		if(client_sock > 0)
		{
			short msg = htons(OP_SHUTDOWN);
			send(client_sock, &msg, 2, 0);
			close(client_sock);
		}
	}
	else if(s == SIGPIPE)
	{
		printf("sig pipe\n");
		// todo: restart lh_w.exe
		if(run_client())
			return;
		printf("can not launch lh_w.exe.\n");
	}
	
	exit(0);
}

int main( int   argc,
          char *argv[] )
{
	if(argc == 2)
	{
		char *key = argv[1];
		if(strcmp(key, "ctrl") == 0)
			key_mask = GDK_CONTROL_MASK;
		else if(strcmp(key, "alt") == 0)
			key_mask = GDK_MOD1_MASK;
		else if(strcmp(key, "shift") == 0)
			key_mask = GDK_SHIFT_MASK;
		else if(strcmp(key, "win") == 0)
			key_mask = GDK_MOD4_MASK;
		else if(strcmp(key, "-h") == 0 || strcmp(key, "--help") == 0)
		{
			printf("LingoesHelper: enalbes Lingoes to capture and translate text in linux.\nUsage: lh_l [ctrl|alt|shift|win]\nAuthor: ausun. email: ausun1@163.com\n");
			return 0;
		}
	}
	
	signal(SIGINT, sig);
	signal(SIGTERM, sig);
	signal(SIGPIPE, sig);	

	/*size = sizeof(struct sockaddr_in);
	client_sock = accept(server_sock,  (struct sockaddr *) &clientaddr, &size);
	if(client_sock < 0)
	{
		printf("accept failed.\n");
		close(server_sock);
		return 1;
	}*/
	if(!run_client())
	{
		printf("can not launch lh_w.exe.\n");
		return 1;
	}
	//printf("client connected.\n");
	//send(client_sock, "hello", 6, 0);
	
	
	gtk_init (&argc, &argv);

	g_sel = new Selection();
	if(!g_sel)
	{
		printf("create selection instance failed.\n");
		short msg = htons(OP_SHUTDOWN);
		send(client_sock, &msg, 2, 0);
		close(client_sock);
		return 1;
	}
	g_sel->Init();
	g_sel->start();

	gtk_main ();

	g_sel->stop();
	short msg = htons(OP_SHUTDOWN);
	send(client_sock, &msg, 2, 0);
	close(client_sock);
	
	delete g_sel;

	return 0;
}
